-- written by DaddelZeit
-- DO NOT USE WITHOUT PERMISSION

local M = {}
M.doDebug = true

local triCache = {}
local curCache = {}
local playerVeh = be:getPlayerVehicle(0)

local soundSmoother = newTemporalSmoothing(4,4)
local lastState = 0
local defaultFilerVal = 1

local driverCamNodeId = 0

local hits = 0
local totalCasts = 0
local curCastNr = 1

local castDirs = {
  onlyUp = {
    {vec3(0,-0.35,1)},
    {vec3(0,0,1)},
    {vec3(0,0.35,1)}
  },
  upAndSides = {
    {vec3(0,-0.45,1)},
    {vec3(0,0,1)},
    {vec3(-1,0,0)},
    {vec3(-1,-0.3,0)},
    {vec3(-1,0.45,0)},
    {vec3(1,-0.3,0)},
    {vec3(1,0,0)},
    {vec3(1,0.45,0)},
  }
}
local setting = "upAndSides"

local function castActual(origin, dir, tri)
  local cast = intersectsRay_Triangle(
    origin,
    dir,
    playerVeh:getNodeAbsPosition(tri[1]),
    playerVeh:getNodeAbsPosition(tri[2]),
    playerVeh:getNodeAbsPosition(tri[3])
  )

  --[[
  if cast > 0 and cast < (max or math.huge) then
    if M.doDebug then debugDrawer:drawSphere(origin+dir*cast, 0.1, ColorF(1,0,0,1)) end
    return true
  end
  ]]
  return cast > 0 and cast < math.huge
end

local function castPhysicsRays(origin, dir, cache)
  for i=1, #cache do
    local cast = castActual(origin, dir, cache[i])
    if cast then return true end
  end
end

local function isWithinRadius(cameraName, camPos, veh, vdata, radius)
  if not vdata.cameras[cameraName] then return false end
  nodePos:set(veh:getNodeAbsPositionXYZ(vdata.cameras[cameraName].camNodeID))
  return nodePos:squaredDistance(camPos) < radius * radius
end

local function onUpdate(dtSim)
  if not driverCamNodeId then
    M.onUpdate = nop
    return
  end
  if bullettime.getPause() or not core_camera or not core_camera.isCameraInside(0, core_camera.getPosition()) then
    return
  end

  curCastNr = curCastNr<#curCache and curCastNr + 1 or 1
  local curLocalCastDirs = castDirs[setting][curCastNr]
  local playerRot = quat(playerVeh:getRefNodeRotation())
  for i=1, #curLocalCastDirs do
    local outsideView = castPhysicsRays(playerVeh:getNodeAbsPosition(driverCamNodeId), curLocalCastDirs[i]:rotated(playerRot), curCache[curCastNr][i])
    totalCasts = totalCasts + 1

    if outsideView then
      hits = hits + 1
    end
  end

  if curCastNr == 1 then
    local fac = hits/totalCasts
    fac = (1-(1-fac)^0.5)
    soundSmoother:get(fac*defaultFilerVal, dtSim)

    hits = 0
    totalCasts = 0

    if soundSmoother.state ~= lastState then
      lastState = soundSmoother.state
      core_sounds.cabinFilterStrength = soundSmoother.state
    end
  end
end

local function cacheRay(vdata, origin, direction)
  local res = {}
  for _,v in pairs(vdata.triangles) do
    if v.triangleType ~= 2 then
      local triPos = (vdata.nodes[v.id1].pos+vdata.nodes[v.id2].pos+vdata.nodes[v.id3].pos)/3

      if triPos:squaredDistanceToLine(origin, origin+direction) < 0.4 and direction:dot((triPos - origin):normalized()) > 0.5 then
        table.insert(res, {v.id1, v.id2, v.id3})
      end
    end
  end
  return res
end

local function cacheMode(vdata, directions, driverCamNodePos)
  local completeTable = {}
  for index, rays in ipairs(directions) do
    local res = {}
    for dirIndex, dir in ipairs(rays) do
      res[dirIndex] = cacheRay(vdata, driverCamNodePos, dir)
    end
    completeTable[index] = res
  end
  return completeTable
end

local function updateVehicle(newId)
  playerVeh = be:getObjectByID(newId)

  local vdata = core_vehicle_manager.getVehicleData(newId)
  if not vdata then return newId end
  vdata = vdata.vdata

  for _,v in ipairs(vdata.nodes) do
    if v.name == "driver" then
      driverCamNodeId = v.cid
      break
    end
  end

  defaultFilerVal = vdata.sounds and vdata.sounds.cabinFilterCoef or 0.8
  if driverCamNodeId == 0 or not vdata.nodes[driverCamNodeId] then
    driverCamNodeId = nil
    return
  end

  local driverCamNodePos = vec3(vdata.nodes[driverCamNodeId].pos)

  triCache = {}
  for mode, rays in pairs(castDirs) do
    triCache[mode] = cacheMode(vdata, rays, driverCamNodePos)
  end
  curCache = triCache[setting]

  soundSmoother:set(defaultFilerVal)
end

local function onVehicleSwitched(_, newId)
  updateVehicle(newId)
end

local function reload()
  extensions.reload("zeit_audioRayCast")
  zeit_audioRayCast.onVehicleSwitched(0, be:getPlayerVehicleID(0))
end

M.onUpdate = onUpdate
M.onVehicleSwitched = onVehicleSwitched
M.reload = reload

return M
